//----------------------------------------------------------------------------
//
// LFSR functions with x86 optimizations, by Scott Duplichan
// tested with Microsoft and mingw C compilers
//
//----------------------------------------------------------------------------
#include "project.h"
#include <windows.h>
//----------------------------------------------------------------------------

#if defined (__GNUC__) && !defined (_xgetbv)
#define _XCR_XFEATURE_ENABLED_MASK 0
static uint64_t _xgetbv (uint32_t ecx)
   {
   uint32_t eax, edx;
   __asm__ ("xgetbv" : "=a" (eax), "=d" (edx) : "c" (ecx)); 
   return ((uint64_t) edx << 32) | eax;
   }
#endif

//----------------------------------------------------------------------------

#if defined (__GNUC__)

static void cpuid (int CPUInfo[4], int InfoType)
  {
   __asm__ __volatile__ (
       "xorl %%ecx, %%ecx;"  // Clear ecx before executing cpuid
       "cpuid"
       : "=a" (CPUInfo [0]), "=b" (CPUInfo [1]), "=c" (CPUInfo [2]), "=d" (CPUInfo [3])
       : "a" (InfoType));
  }
#else

static void cpuid (int CPUInfo[4], int InfoType)
  {
  __cpuid (CPUInfo, InfoType); // Microsoft compiler clears ecx before executing cpuid
  }

#endif

//----------------------------------------------------------------------------
// check for processor support for avx instructions

static int avxAvailable (void)
   {
   int regs [4];
   uint64_t xcrFeatureMask;
   cpuid (regs, 1);
   if ((regs [2] & 0x18000000) != 0x18000000) return 0;

   xcrFeatureMask = _xgetbv (_XCR_XFEATURE_ENABLED_MASK);
   if ((xcrFeatureMask & 6) != 6) return 0;
   return 1;
   }

//----------------------------------------------------------------------------
// check for processor support for avx2 instructions

static int avx2Available (void)
   {
   int regs [4];

   if (!avxAvailable ()) return 0;
   cpuid (regs, 7);
   if ((regs [1] & 0x20) == 0) return 0;
   return 1;
   }

//-----------------------------------------------------------------------------
//
//  queryPerformanceCounter - similar to QueryPerformanceCounter, but returns
//                            count directly.

uint64_t queryPerformanceCounter (void)
   {
   LARGE_INTEGER int64;
   QueryPerformanceCounter (&int64);
   return int64.QuadPart;
   }

//-----------------------------------------------------------------------------
//
// queryPerformanceFrequency - same as QueryPerformanceFrequency, but returns  count direcly.

uint64_t queryPerformanceFrequency (void)
   {
   LARGE_INTEGER int64;

   QueryPerformanceFrequency (&int64);
   return int64.QuadPart;
   }

//----------------------------------------------------------------------------

char *cyclesText (uint64_t elapsedCount, uint64_t cpuSpeed, uint64_t loops)
   {
   static char buffer [80];
   double cycles;

   if (cpuSpeed == 0)
      {
      cycles = (double) elapsedCount / queryPerformanceFrequency () * 1000000000 / loops;
      sprintf (buffer, "%7.1f ns", cycles);
      return buffer;
      }

   cycles = (double) elapsedCount / queryPerformanceFrequency () * cpuSpeed / loops;
   sprintf (buffer, "%7.1f cycles", cycles);
   return buffer;
   }

//----------------------------------------------------------------------------

int main (int argc, char *argv [])
   {
   // Primitive polynomials, initial patterns, and expected final pattern
   // for each lfsr size. Example for finding primitive polynomial (128 bits):
   //    C:>ppsearch hex bits=128 weight=127
   //    17FDFFFFFFFFFFFFFFFFFFFFFFFFFFFFF
   // Example for calculating the final state (128 bits, 10000000 iterations):
   //    C:>gf2util display=hex modpower,0x17FDFFFFFFFFFFFFFFFFFFFFFFFFFFFFF,10b,10000000d
   //    0xE9E9FF260A0BF0694B4DC45E248DFFC5
   //    C:>gf2util display=hex modmul,0x17FDFFFFFFFFFFFFFFFFFFFFFFFFFFFFF,0xE9E9FF260A0BF0694B4DC45E248DFFC5,0x12345678123456781234567812345678
   //    0x5FADA3A914EB5CE3027F870C421B1A6C

   static TEST_DATA testData = {
      // mask32,mask64, mask128, mask256, mask512, mask1024, mask2048, mask4096
      {0x000000005FFFFFFF, 0},
      {0xBEFFFFFFFFFFFFFF, 0},
      {0xFFFFFFFFFFFFFFFF, 0x7FDFFFFFFFFFFFFF},
      {0xFFFFFFFFFFFFFFFF, 0xFFFFFFFFFFFFFFFF, 0xFFFFFFFFFFFFFFFF, 0xFEF7FFFFFFFFFFFF},
      {0xFFFFFFFFFFFFFFFF, 0xFFFFFFFFFFFFFFFF, 0xFFFFFFFFFFFFFFFF, 0xFFFFFFFFFFFFFFFF, 0xFFFFFFFFFFFFFFFF, 0xFFFFFFFFFFFFFFFF, 0xFFFFFFFFFFFFFFFF, 0xFFFFF9FFFFFFFFFF},
      {0xFFFFFFFFFFFFFFFF, 0xFFFFFFFFFFFFFFFF, 0xFFFFFFFFFFFFFFFF, 0xFFFFFFFFFFFFFFFF, 0xFFFFFFFFFFFFFFFF, 0xFFFFFFFFFFFFFFFF, 0xFFFFFFFFFFFFFFFF, 0xFFFFFFFFFFFFFFFF, 0xFFFFFFFFFFFFFFFF, 0xFFFFFFFFFFFFFFFF, 0xFFFFFFFFFFFFFFFF, 0xFFFFFFFFFFFFFFFF, 0xFFFFFFFFFFFFFFFF, 0xFFFFFFFFFFFFFFFF, 0xFFFFFFFFFFFFFFFF, 0xDFFBFFFFFFFFFFFF},
      {0xFFFFFFFFFFFFFFFF, 0xFFFFFFFFFFFFFFFF, 0xFFFFFFFFFFFFFFFF, 0xFFFFFFFFFFFFFFFF, 0xFFFFFFFFFFFFFFFF, 0xFFFFFFFFFFFFFFFF, 0xFFFFFFFFFFFFFFFF, 0xFFFFFFFFFFFFFFFF, 0xFFFFFFFFFFFFFFFF, 0xFFFFFFFFFFFFFFFF, 0xFFFFFFFFFFFFFFFF, 0xFFFFFFFFFFFFFFFF, 0xFFFFFFFFFFFFFFFF, 0xFFFFFFFFFFFFFFFF, 0xFFFFFFFFFFFFFFFF, 0xFFFFFFFFFFFFFFFF, 0xFFFFFFFFFFFFFFFF, 0xFFFFFFFFFFFFFFFF, 0xFFFFFFFFFFFFFFFF, 0xFFFFFFFFFFFFFFFF, 0xFFFFFFFFFFFFFFFF, 0xFFFFFFFFFFFFFFFF, 0xFFFFFFFFFFFFFFFF, 0xFFFFFFFFFFFFFFFF, 0xFFFFFFFFFFFFFFFF, 0xFFFFFFFFFFFFFFFF, 0xFFFFFFFFFFFFFFFF, 0xFFFFFFFFFFFFFFFF, 0xFFFFFFFFFFFFFFFF, 0xFFFFFFFFFFFFFFFF, 0xFFFFFFFFFFFFFFFF, 0xFFFFFFFFFFBFFFFB},
      {0xFFFFFFFFFFFFFFFF, 0xFFFFFFFFFFFFFFFF, 0xFFFFFFFFFFFFFFFF, 0xFFFFFFFFFFFFFFFF, 0xFFFFFFFFFFFFFFFF, 0xFFFFFFFFFFFFFFFF, 0xFFFFFFFFFFFFFFFF, 0xFFFFFFFFFFFFFFFF, 0xFFFFFFFFFFFFFFFF, 0xFFFFFFFFFFFFFFFF, 0xFFFFFFFFFFFFFFFF, 0xFFFFFFFFFFFFFFFF, 0xFFFFFFFFFFFFFFFF, 0xFFFFFFFFFFFFFFFF, 0xFFFFFFFFFFFFFFFF, 0xFFFFFFFFFFFFFFFF, 0xFFFFFFFFFFFFFFFF, 0xFFFFFFFFFFFFFFFF, 0xFFFFFFFFFFFFFFFF, 0xFFFFFFFFFFFFFFFF, 0xFFFFFFFFFFFFFFFF, 0xFFFFFFFFFFFFFFFF, 0xFFFFFFFFFFFFFFFF, 0xFFFFFFFFFFFFFFFF, 0xFFFFFFFFFFFFFFFF, 0xFFFFFFFFFFFFFFFF, 0xFFFFFFFFFFFFFFFF, 0xFFFFFFFFFFFFFFFF, 0xFFFFFFFFFFFFFFFF, 0xFFFFFFFFFFFFFFFF, 0xFFFFFFFFFFFFFFFF, 0xFFFFFFFFFFFFFFFF, 0xFFFFFFFFFFFFFFFF, 0xFFFFFFFFFFFFFFFF, 0xFFFFFFFFFFFFFFFF, 0xFFFFFFFFFFFFFFFF, 0xFFFFFFFFFFFFFFFF, 0xFFFFFFFFFFFFFFFF, 0xFFFFFFFFFFFFFFFF, 0xFFFFFFFFFFFFFFFF, 0xFFFFFFFFFFFFFFFF, 0xFFFFFFFFFFFFFFFF, 0xFFFFFFFFFFFFFFFF, 0xFFFFFFFFFFFFFFFF, 0xFFFFFFFFFFFFFFFF, 0xFFFFFFFFFFFFFFFF, 0xFFFFFFFFFFFFFFFF, 0xFFFFFFFFFFFFFFFF, 0xFFFFFFFFFFFFFFFF, 0xFFFFFFFFFFFFFFFF, 0xFFFFFFFFFFFFFFFF, 0xFFFFFFFFFFFFFFFF, 0xFFFFFFFFFFFFFFFF, 0xFFFFFFFFFFFFFFFF, 0xFFFFFFFFFFFFFFFF, 0xFFFFFFFFFFFFFFFF, 0xFFFFFFFFFFFFFFFF, 0xFFFFFFFFFFFFFFFF, 0xFFFFFFFFFFFFFFFF, 0xFFFFFFFFFFFFFFFF, 0xFFFFFFFFFFFFFFFF, 0xFFFFFFFFFFFFFFFF, 0xFFFFFFFFFFFFFFFF, 0xFFDFFBFFFFFFFFFF},
      // expected32, expected64, expected128, expected256, expected512, expected1024, expected2048, expected4096
      {0x000000006B6A4180, 0},
      {0xF4AEA9BEF253E1BD, 0},
      {0x027F870C421B1A6C, 0x5FADA3A914EB5CE3},
      {0x7DE9956F4B0E15D4, 0x37133E33E0F0B44B, 0xC66BF8D0F77C5A76, 0xF0CF004432305C06},
      {0xC4F56E7A6881008B, 0x122FAF8C2A555AFF, 0x8E65D78ED8575FAB, 0xB29404C4329D2318, 0x02D4A81D2806C261, 0x18B53C30623BE1DD, 0x7BD6D4C4C0E4FC86, 0x76AE494E70229E83},
      {0x42E993D623400A6C, 0x9D6E30191FE34255, 0xA243A5D307F4875A, 0x1280BF2819826F71, 0xABA87DD31676F9BB, 0xE23A0559619F7C5E, 0x4114BCFC1AE9AD3D, 0x28ACFDF933AE891E, 0xC6BFAA053D0C3D6B, 0x1D988D28C54A96C2, 0xCA76EBC45FE5D5B3, 0x76A5EACF935BF665, 0xB0B7B25FE2DAD2DA, 0x41CC7E45FCBCC4D5, 0xF88F42B8C30E7C6F, 0x5489917C0FA8FBE5},
      {0xEDF50A41C3AA30D9, 0xD81779664A8E9626, 0x2615DEDF721E4B18, 0x111B5AC50E8ACBFC, 0x1FB30CFF63D8BA3F, 0x35C699ADB2A529B1, 0x8CD37D861B201647, 0xC85246280934A18A, 0xF7763B8F2665244D, 0x89B1A86C6B527B8F, 0xDD9CFAAF4E0CBD7A, 0x376B47FDA47E34A8, 0x0D36FEF8D8E53A0B, 0x8FCC70D902B5CC5A, 0x6B8281A4F1B9C32A, 0xA781A62E276F83CD, 0x02B8651DE45E972A, 0xBDAFCF231E78E7E4, 0x97DD7A7FDB6F78AC, 0x75F07F5BEA3C46EA, 0x4A0E1F7CD0620128, 0x560D57383D587C60, 0x53749CF8954F9C48, 0xEF1399D8E6315146, 0x121A46767D8AE39A, 0x87A1EDC24D9EDBAF, 0x052009E82EB83AE0, 0xCFE76681ABB16597, 0xDFF37AF18C2AD737, 0xF459FBDC54A7A8E5, 0x401655EC3BA28D4B, 0xF069ABBD32F2FCB8},
      {0xCC50267C250D7264, 0x7637D27CD6CB6191, 0xFA873B6DDFA93197, 0xDB7F33217C93CA2E, 0x944CE031D00BDBD6, 0x067AAE28AF2C7F86, 0xE8439285F729D46E, 0xAD22C5EB6B36BEA4, 0xA79A0F5442D244CF, 0x9EAC63A1CCA67B7C, 0x3708F0407EDD04F1, 0x5165AD0A593A7421, 0xF8F1DD403BB4953B, 0x613C49136431866E, 0x0D6DDC9C305957E4, 0xFC8C37142024F4DD, 0xF71F62033F31E6F6, 0xD36EE0FFCE17E4ED, 0x13E984791E287A67, 0x87609166E5A64BC2, 0x922EAFC6CBF90EEB, 0xC4E1819F0236F072, 0x5BAB3638D7A1931F, 0xA922A3840516564B, 0xD6FD6C4E274D4B9B, 0x820BF9DDE606A3A2, 0x0A6798FED8CB0263, 0x2689B80286CCCC70, 0xF0A9B773A4DDAD7E, 0x499380162F73EE5C, 0xBA55EC2DEF6F30F1, 0x2DD73CCD9FA60631, 0x1D8320CDD70D7CD6, 0x9FA1CF072C0A9EB5, 0x4482CA02B2DB3208, 0xB438160208F890A8, 0xFD4D394830B12D30, 0x23A123DFDA5A7F6E, 0x8EE9333F1111B0BE, 0x5AD3CE4BA0F690FB, 0xE472125F5F4CF3AB, 0x920250248EF5C648, 0x1025B9A981BC8BBE, 0xC10C8349D450C658, 0xB60CEB28BD3D3BD7, 0x94D1BB9E6A4937F0, 0x53D916B4A07B6EE3, 0x2DA7CABF037B7741, 0xB70BE1C2BE79AD7B, 0x77CAB165CC59D434, 0x32060FAE2731F4EB, 0xA91C5C8B84960505, 0xF4DD532411336169, 0x147E9CE6CB604E0A, 0xC729D67E25000B19, 0x697563CD1AD95987, 0xBB2A407F23F1E77F, 0xF30AA46020ECC6A7, 0x948E502B529C6DCF, 0x555B078EB62DDE53, 0xB957995A76FC925D, 0x2B5982822A24C721, 0xEA52C232E2C2C22F, 0xBD787D66AD70BAC1},

      // initial
      {0x1234567812345678, 0x1234567812345678, 0x1234567812345678, 0x1234567812345678, 0x1234567812345678, 0x1234567812345678, 0x1234567812345678, 0x1234567812345678, 0x1234567812345678, 0x1234567812345678, 0x1234567812345678, 0x1234567812345678, 0x1234567812345678, 0x1234567812345678, 0x1234567812345678, 0x1234567812345678, 0x1234567812345678, 0x1234567812345678, 0x1234567812345678, 0x1234567812345678, 0x1234567812345678, 0x1234567812345678, 0x1234567812345678, 0x1234567812345678, 0x1234567812345678, 0x1234567812345678, 0x1234567812345678, 0x1234567812345678, 0x1234567812345678, 0x1234567812345678, 0x1234567812345678, 0x1234567812345678, 0x1234567812345678, 0x1234567812345678, 0x1234567812345678, 0x1234567812345678, 0x1234567812345678, 0x1234567812345678, 0x1234567812345678, 0x1234567812345678, 0x1234567812345678, 0x1234567812345678, 0x1234567812345678, 0x1234567812345678, 0x1234567812345678, 0x1234567812345678, 0x1234567812345678, 0x1234567812345678, 0x1234567812345678, 0x1234567812345678, 0x1234567812345678, 0x1234567812345678, 0x1234567812345678, 0x1234567812345678, 0x1234567812345678, 0x1234567812345678, 0x1234567812345678, 0x1234567812345678, 0x1234567812345678, 0x1234567812345678, 0x1234567812345678, 0x1234567812345678, 0x1234567812345678, 0x1234567812345678},
   };
   uint64_t cpuSpeed = 0;
   int avx = avxAvailable ();
   int avx2 = avx2Available ();
   int loops = 10000000;
   DWORD priority;

   printf ("fastlfsr v1.3\n");
   if (argc == 2) sscanf (argv [1], "%I64u", &cpuSpeed);

   // raise our priority to increase measurement accuracy
   priority = GetPriorityClass (GetCurrentProcess ());
   SetPriorityClass (GetCurrentProcess (), REALTIME_PRIORITY_CLASS);

   lfsrGpr (&testData, loops, cpuSpeed);
   if (avx) lfsrXmm (&testData, loops, cpuSpeed);
   if (avx2) lfsrYmm (&testData, loops, cpuSpeed);

   // restore priority
   SetPriorityClass (GetCurrentProcess (), priority);

   return 0;
   }

//----------------------------------------------------------------------------
